#pragma once

#include <LSimpleThread.h>

#ifndef __hydra__
#define __hydra__

///////////////////////////////////////////////////////////////
//	Hydra classes are Copyright 1997 by Rainer Brockerhoff.
//
//	Submitted as part of the paper :
//	"On Having Two (or More) Heads, or : Real-time Programming in Spite of
//		the MacOS", for MacHack 1997
//
//	You may use these classes in your application at no cost;
//	however, the application's "About" box should repeat my
//	copyright line as above. Be sure to write my name correctly ;-)
//	I'd also appreciate getting a demo copy of your application.
//	These classes have not been exhaustively tested in this form
//	and all use is at your own risk.
//
//	Contact me at <rainer@machome.com.br> or
//	http://www.machome.com.br/delta/
//
//	Stylistic note :
//	Astute older colleagues will no doubt deduce that in my misspent youth I:
//	1) learned FORTRAN	(short variable names; local variables called i,j,k)
//	2) learned PL/I		(using }; everywhere)
//	3) used punch cards (using {} even when not needed, to more easily insert
//		printf debug lines)
//	4) had too little contact with other programmers and so failed to learn
//		proper indentation, naming or bracing standards.
//	They're probably right on all counts... :-)
//
//	I'm assuming CodeWarrior 12 (or Pro 1) is being used.

enum	{ oneSecond = 1000 };		// convenient constant for use with the
									// MilliSecond() and Sleep() calls

class	HydraApplication;			// forward declarations
class	HydraWindow;
class	Hydra;

///////////////////////////////////////////////////////////////
// Declare your application as a subclass of HydraApplication.
// This will handle all necessary initialization and install
// the special threading features.
//
// Be careful when overriding our internal functions!
//
class	HydraApplication : public LApplication {
public:
							HydraApplication();
	virtual 				~HydraApplication();
protected:
	class	HydraMenuBar : public LMenuBar {			// our private menu bar
	public:												// this is needed
							HydraMenuBar(ResIDT inMBARid);// to install our
	protected:											// MBarHook procedure
		virtual				~HydraMenuBar();
		virtual	CommandT	MenuCommandSelection(const EventRecord &inMouseEvent,Int32 &outMenuChoice) const;
	};
	virtual void			MakeMenuBar();		// makes private menu bar
private:
	virtual void			ProcessNextEvent();
	virtual void			DispatchEvent(const EventRecord &inMacEvent);
	virtual void			TryToDispatch();
	volatile EventMask		mEventMask;			// see "Hydra.cp" about this
friend class Hydra;
};

///////////////////////////////////////////////////////////////
// Declare your real-time windows as subclasses of HydraWindow.
// This will handle all necessary initialization and most extra
// functionality.
// You don't need to do this for windows (like LDialogBoxes) that are
// opened and updated by the main thread.
// Notice that the formerly protected method FocusDraw() is now public!
//
// Be careful when overriding our internal functions!
//
class	HydraWindow : public LWindow {
public:
	enum { class_ID = 'Hwin' };					// use this in Constructor
	static HydraWindow*		CreateWindow(ResIDT inWindowID,
								LCommander* inSuperCommander);
							HydraWindow();
							HydraWindow(const SWindowInfo &inWindowInfo);
							HydraWindow(ResIDT inWINDid,Uint32 inAttributes,
								LCommander* inSuper);
							HydraWindow(LStream* inStream);
	virtual					~HydraWindow();
	virtual Boolean			FocusDraw(LPane* inSubPane=NULL);
protected:
	virtual void			ClickInDrag(const EventRecord &inMacEvent);
	virtual void			ClickInGrow(const EventRecord &inMacEvent);
};

///////////////////////////////////////////////////////////////
// Use Hydra as a stack-based state saver :
//	{
//		Hydra saver;		// this saves Grafport
//		...
//		::SetPort(whatever);
//		... // draw elsewhere
//		...
//		Hydra::Reset();		// restore saved GrafPort
//		... // draw in original Grafport
//		...
//		::SetPort(something);
//		... // draw elsewhere
//	};						// this restores GrafPort and handles LView cache.
//
// Other public functions : (see "Hydra.cp" for more info)
//	Hydra::MilliSeconds() : returns a millisecond counter. Zero is (roughly)
//								the application's startup time.
//	Hydra::Yield() : call this instead of LThread::Yield() to yield time to
//								other threads.
//	Hydra::Sleep() : call this instead of LThread::Sleep() to sleep for a
//								while, yielding time to other threads.
//	RedrawUnderRegion() : call this for immediate redraw of the HydraWindows
//								under the parameter region.
//	SetWNEFrequency() : call this to determine how many times per second you
//								want ::WaitNextEvent() to be called.
//
// The other functions are for internal use only. See "Hydra.cp" for more info.
//
//	This source file is overcommented, for pedagogical purposes only.
//
#pragma warn_hidevirtual off		// need this to suppress warning messages
template class TArray<WindowPtr>;	// declare an array template
#pragma warn_hidevirtual reset

class	Hydra {
public:
							Hydra();		// constructor saves GrafPort
	virtual					~Hydra();		// destructor calls Reset()
	void					Reset();		// restores GrafPort and handles
											//	LView cache
	static void				Yield(const LThread* yieldTo=NULL);	// yield to
																//	other threads
	static UInt32			MilliSeconds();				// milliseconds from
														//	application startup
	static void				Sleep(Int32 milliSeconds);	// sleeps while letting
														//	others run
	static RgnHandle		GetMenuRegion();// accessor for menu region
	static void				RedrawUnderRegion(RgnHandle theRgn=NULL);// redraws
													//	windows under region
	static void				SetWNEFrequency(short perSecond);// sets frequency
													//	for main event loop
	static Boolean			ShieldMenu(LView* theView);// shield menu from view
private:					// everything below this is used internally
							// see "Hydra.cp" for more information
	class	HydraDragAndDrop : public LDropArea {
	public:
							HydraDragAndDrop(WindowPtr theWindow=NULL);
		virtual				~HydraDragAndDrop();
	protected:
		static pascal OSErr	DragTracking(DragTrackingMessage inMessage,WindowPtr inMacWindow,void *inRefCon,DragReference inDragRef);
		virtual Boolean		PointInDropArea(Point inGlobalPt);
	};
	static void				RememberWindow(HydraWindow* theWindow);
	static void				ForgetWindow(HydraWindow* theWindow);
	static void				Initialize(HydraApplication* inApp);
	static void				ShutDown();
	static pascal short		MBarHook(Rect* menuRect);
	static pascal Boolean	StillDownPatch();
	static pascal Boolean	WaitMouseUpPatch();
	GrafPtr					mGrafPtr;
	static struct Storage {					// local storage for our stuff
		TArray<WindowPtr> 	sWindows;		// list of shielded windows (their GrafPorts)
		TArray<WindowPtr>	sHydraWindows;	// list of HydraWindows (their GrafPorts)
		unsigned long long	sStartedAt;		// start time (from ::MicroSeconds())
		MBarHookUPP			sMBarHookUPP;	// our MBarHook
		DragTrackingHandlerUPP sOldDragTracking;// PowerPlant's drag tracker
		UniversalProcPtr	sOldStillDown;	// remembers old ::StillDown() address
		UniversalProcPtr	sOldWaitMouseUp;// remembers old ::WaitMouseUp() address
		HydraApplication*	sTheApp;		// points at application object
		RgnHandle			sMenuRegion;	// region covered by current menu
		UInt32 volatile		sMainTime;		// next time main thread should run
		Int32				sDelay;			// delay between main thread runs
		Boolean volatile	sInWaitMouseUp;	// true if inside ::WaitMouseUp()
	} sHydraStorage;
friend class HydraApplication;
friend class HydraWindow;
friend class Hydra::HydraDragAndDrop;
};

#endif

///////////////////////////////////////////////////////////////
//	end of Hydra.h


