/****************************************************************************************
	XThreads.c

	Copyright  1997 United Software Express. All rights reserved.

	Commenter	Date				Comment
	---------	-----------------	-----------------------------------------------------
	wolfie		Mon, Mar 3, 1997	Created.
	wolfie		Sun, May 25, 1997	Word of caution: XThreads allocates new handles
									when it runs out of space but never releases
									the handles. I could see where if you allocate a
									thousand threads and dump them all this could be
									an issue. However, 1024 threads would only take
									12K of memory, which leads me to think if you're
									allocating a thousand threads, 12K isn't your
									biggest concern.
	
****************************************************************************************/

#include	"XThreadsQueue.h"

XThreadQueue		gUnassignedQueue, gIneligibleQueue, gEligibleQueue, gPriorityQueue;
XThreadElemPtr		gCurrentThread;

const	short		kGrowCount	=	32;
const	short		kGrowSize	=	kGrowCount * sizeof( XThreadElem );

pascal	void		XThreadDebuggerNewThread( ThreadID threadCreated );
pascal	void		XThreadDebuggerDisposeThread( ThreadID threadDeleted );
pascal	ThreadID	XThreadDebuggerThreadScheduler( SchedulerInfoRecPtr schedulerInfo );

#pragma mark	--XThreads Interface Functions--

/**************************
 *
 *	InitXThreads
 *
 */

OSErr		InitXThreads()
{
	long			response;
	ThreadID		mainThreadID;
	XThreadElemPtr	mainThreadElem;
	OSErr			err;
	
	/*	Test if Thread Manager exists */
	err = Gestalt( gestaltThreadMgrAttr, &response );
	if( !err )
		err = ((( response >> gestaltThreadMgrPresent ) & 1 ) != 0 ) ? noErr : 1;
	
	/*	Initialize the queues */
	if( !err ) {
		gUnassignedQueue.head = gUnassignedQueue.tail = gUnassignedQueue.mark = nil;
		gIneligibleQueue.head = gIneligibleQueue.tail = gIneligibleQueue.mark = nil;
		gEligibleQueue.head = gEligibleQueue.tail = gEligibleQueue.mark = nil;
		gPriorityQueue.head = gPriorityQueue.tail = gPriorityQueue.mark = nil;
	}
		
	/*	Add the main thread to the list of eligible threads and mark it as the current thread */
	if( !err )
		err = GetCurrentThread( &mainThreadID );
	if( !err ) {
		mainThreadElem = NewThreadElem( mainThreadID );
		err = mainThreadElem ? noErr : memFullErr;
	}
	if( !err ) {
		SetThreadElemQueue( mainThreadElem, &gEligibleQueue );
		gCurrentThread = mainThreadElem;
	}
	
	/*	Install Callbacks */
	if( !err )
		err = SetDebuggerNotificationProcs( XThreadDebuggerNewThread,
											XThreadDebuggerDisposeThread,
											XThreadDebuggerThreadScheduler );
	return( err );
}

/**************************
 *
 *	GetXThreadState
 *
 */

XThreadState	GetXThreadState( ThreadID threadID )
{
	XThreadElemPtr	threadElem;
	XThreadState	threadState = kXThreadEligible;
	
	threadElem = GetThreadElemGivenThreadID( threadID );
	if( threadElem ) {
		if( threadElem->queue == &gIneligibleQueue )
			threadState = kXThreadIneligible;
		else if( threadElem->queue == &gPriorityQueue )
			threadState = kXThreadPriority;
	}
	return( threadState );
}

/**************************
 *
 *	GetCurrentXThreadState
 *
 */

XThreadState	GetCurrentXThreadState()
{
	XThreadState	threadState = kXThreadEligible;
	ThreadID		threadID = kNoThreadID;
	
	/*	Test the validity of our cache */
	if( gCurrentThread ) {
		/*	Our cache is valid -- use it */
		if( gCurrentThread->queue == &gIneligibleQueue )
			threadState = kXThreadIneligible;
		else if( gCurrentThread->queue == &gPriorityQueue )
			threadState = kXThreadPriority;
	} else {
		/*	The cache isn't valid -- do it the hard way */
		(void) GetCurrentThread( &threadID );
		threadState = GetXThreadState( threadID );
	}
	
	return( threadState );
}

/**************************
 *
 *	SetXThreadState
 *
 */

OSErr			SetXThreadState( ThreadID threadID, XThreadState state )
{
	XThreadElemPtr	threadElem;
	XThreadQueuePtr	threadQueue;
	OSErr			err = noErr;
	
	threadElem = GetThreadElemGivenThreadID( threadID );
	if( threadElem ) {
		switch( state ) {
			case kXThreadIneligible:
				threadQueue = &gIneligibleQueue;
				break;
			case kXThreadPriority:
				threadQueue = &gPriorityQueue;
				break;
			default:
				threadQueue = &gEligibleQueue;
				break;
		}
		SetThreadElemQueue( threadElem, threadQueue );
	} else
		err = 1;
	return( err );	
}

/**************************
 *
 *	SetCurrentXThreadState
 *
 */

OSErr			SetCurrentXThreadState( XThreadState state )
{
	XThreadQueuePtr	threadQueue;
	ThreadID		threadID = kNoThreadID;
	OSErr			err = noErr;
	
	/*	Test the validity of our cache */
	if( gCurrentThread ) {
		/*	Our cache is valid -- use it */
		switch( state ) {
			case kXThreadIneligible:
				threadQueue = &gIneligibleQueue;
				break;
			case kXThreadPriority:
				threadQueue = &gPriorityQueue;
				break;
			default:
				threadQueue = &gEligibleQueue;
				break;
		}
		SetThreadElemQueue( gCurrentThread, threadQueue );
	} else {
		/*	Our cache is invalid -- do it the hard way */
		err = GetCurrentThread( &threadID );
		if( !err )
			err = SetXThreadState( threadID, state );
	}
	
	return( err );
}

/****************************************************************************************
*	
*	XThreadsQueue Interface Functions
*	
****************************************************************************************/
#pragma mark	--XThreadsQueue Interface Functions--

/**************************
 *
 *	NewThreadElem
 *
 */

XThreadElemPtr	NewThreadElem( ThreadID threadID )
{
	XThreadElemHandle	elemHandle;
	XThreadElemPtr		elemPtr;
	short				i;
	
	/*	We're out of unassigned elements, allocate another block of elements */
	if( gUnassignedQueue.head == nil ) {
		elemHandle = (XThreadElemHandle) NewHandle( kGrowSize );
		if( elemHandle == nil )
			return( nil );
		HLockHi( (Handle) elemHandle );
		elemPtr = *elemHandle;
		for( i = 0; i < kGrowCount; i++ ) {
			Enqueue( (QElemPtr) &elemPtr[ i ], (QHdrPtr) &gUnassignedQueue );
			elemPtr[ i ].queue = (XThreadQueuePtr) &gUnassignedQueue;
			elemPtr[ i ].threadID = kNoThreadID;
		}
	}
	
	/*	Yank the first element */
	elemPtr = gUnassignedQueue.head;
	Dequeue( (QElemPtr) elemPtr, (QHdrPtr) &gUnassignedQueue );
	elemPtr->queue = nil;
	elemPtr->threadID = threadID;
	return( elemPtr );
}

/**************************
 *
 *	DisposeThreadElem
 *
 */

void			DisposeThreadElem( XThreadElemPtr threadElem )
{
	SetThreadElemQueue( threadElem, &gUnassignedQueue );
}

/**************************
 *
 *	SetThreadElemQueue
 *
 */

void		SetThreadElemQueue( XThreadElemPtr threadElem , XThreadQueuePtr threadQueue )
{
	XThreadElemPtr	oldNext = nil;
	
	/*	Does this element belong to a queue? */
	if( threadElem->queue ) {
		/*	Store this element's next link */
		oldNext = threadElem->next;
		/* Remove this element from the old queue */
		Dequeue( (QElemPtr) threadElem, (QHdrPtr) threadElem->queue );
		/*	Was this element marked? */
		if( threadElem->queue->mark == threadElem ) {
			/*	It was -- mark the next element in the queue. If there isn't a next item,
				mark the first item instead */
			threadElem->queue->mark = oldNext ? oldNext : threadElem->queue->head;
		}
	}
	Enqueue( (QElemPtr) threadElem, (QHdrPtr) threadQueue );
	threadElem->queue = threadQueue;
	if( !threadQueue->mark )
		threadQueue->mark = threadElem;
}

/**************************
 *
 *	GetNextThreadElemInQueue
 *
 */

XThreadElemPtr	GetNextThreadElemInQueue( XThreadQueuePtr threadQueue )
{
	XThreadElemPtr	nextElem = nil;
	
	/*	If no element is marked than mark the first element if there is one. */
	if( !threadQueue->mark )
		threadQueue->mark = threadQueue->head;
	/*	Is there a marked thread? */
	if( threadQueue->mark ) {
		/*	Return the next element and mark the next element */
		nextElem = threadQueue->mark;
		threadQueue->mark = nextElem->next ? nextElem->next : threadQueue->head;
	}
	
	return( nextElem );
}

/**************************
 *
 *	GetThreadElemGivenThreadID
 *
 */

XThreadElemPtr	GetThreadElemGivenThreadID( ThreadID threadID )
{
	XThreadQueuePtr	queues[] = { &gEligibleQueue, &gIneligibleQueue, &gPriorityQueue };
	XThreadElemPtr	value = nil, current;
	short			i;
	
	for( i = 0; !value && i < 3; i++ ) {
		current = queues[ i ]->head;
		
		while( !value && current ) {
			if( current->threadID == threadID )
				value = current;
			else
				current = current->next;
		}
	}
	
	return( value );
}

/****************************************************************************************
*	
*	Thread Manager Debugging Callbacks
*	
****************************************************************************************/
#pragma mark	--Thread Manager Debugging Callbacks--

/**************************
 *
 *	XThreadDebuggerNewThread
 *
 */

pascal	void	XThreadDebuggerNewThread( ThreadID threadCreated )
{
	XThreadElemPtr	threadElem;
	
	threadElem = NewThreadElem( threadCreated );
	if( threadElem )
		SetThreadElemQueue( threadElem, &gEligibleQueue );
}

/**************************
 *
 *	XThreadDebuggerDisposeThread
 *
 */

pascal	void	XThreadDebuggerDisposeThread( ThreadID threadDeleted )
{
	XThreadElemPtr	threadElem;
	
	threadElem = GetThreadElemGivenThreadID( threadDeleted );
	if( threadElem )
		DisposeThreadElem( threadElem );
}

/**************************
 *
 *	XThreadDebuggerThreadScheduler
 *
 */

pascal	ThreadID	XThreadDebuggerThreadScheduler( SchedulerInfoRecPtr schedulerInfo )
{
	ThreadState		threadState = kStoppedThreadState;
	XThreadElemPtr	threadElem;
	OSErr			err = noErr;
	
	/*	Test if there's a priority thread */
	threadElem = GetNextThreadElemInQueue( &gPriorityQueue );
	if( threadElem )
		/*	Place the priority thread into the eligible group */
		SetThreadElemQueue( threadElem, &gEligibleQueue );
	/*	Test if there was a priority thread or if there's a "suggested" thread */
	if( !threadElem && schedulerInfo->SuggestedThreadID == kNoThreadID ) {
		/*	Look for a thread that's state is Thread Manager-ready. */
		while( !err && threadState == kStoppedThreadState ) {
			threadElem = GetNextThreadElemInQueue( &gEligibleQueue );
			err = threadElem ? GetThreadState( threadElem->threadID, &threadState ) : 1;
		}
	}
	/*	Cache the current thread */
	gCurrentThread = threadElem;
	/*	If we didn't find a thread to schedule, just choose the "suggested" thread. */
	return( threadElem ? threadElem->threadID : schedulerInfo->SuggestedThreadID );
}