/**********************************************************************************
	FILE: HandleIPLSCalls.cp
	Description: These routines get information from the IPLS Server.  They open an
		OT endpoint, send the request, and get back the lines of data.
		
		The OT routines need an Idler since they will be waiting for data, but we
		don't want to return from the routines until we are done.  It would be
		possible to do these through polling, but this is both simpler and more
		understandable.

	1997, CE Software, Inc.  All rights reserved.
	P.O. Box 65580
	West Des Moines, Iowa 50265 U.S.A.

***********************************************************************************/

#include <string.h>

#include "IPLSBROTEndpoint.h"
#include "HandleIPLSCall.h"
#include "ICAPI.h"
#include "QuickTalker.h"

BeachBall gTheBall;
IPLSIdler gTheIdler;
IPLSBROTEndPoint* gTheConnection = nil;


static void DoCancel(void);	//forward reference so our idler can cancel our connection

#define kDefaultWNEDelay	15
#define kDefaultIdlerPeriod	15
IPLSIdler::IPLSIdler (void)		//default constructor.  Tell the Idler base class how often to call yield action
{
	SetPeriod (kDefaultWNEDelay);

	mlWNEDelay = kDefaultIdlerPeriod;
}

void IPLSIdler::YieldAction (void)
{
	EventRecord		sEvent;
	
	if (!gInBackground)
		gTheBall.Idle();		//spin our beach ball cursor
	
	if (WaitNextEvent(everyEvent , &sEvent, mlWNEDelay, nil)) //if escape or command-period pressed, cancel connection
	{
		switch (sEvent.what)
		{
			case kHighLevelEvent:
				AEProcessAppleEvent( &sEvent ) ;
				break;
			case keyDown:
				if (((sEvent.message & charCodeMask)=='.')&&(sEvent.modifiers & cmdKey))
				{
					DoCancel();
				}
				break;
			case kOSEvent:
				if (((sEvent.message >> 24) & 0x0FF) == kSuspendResumeMessage) /* high byte of message */
				{
					gInBackground = (sEvent.message & kResumeMask) == 0;
				}
		}
	}
}

static void DoCancel(void)
{
	if (gTheConnection)
		gTheConnection->Cancel();
}

//
// Parse the port out of the string (if any) and open the connection
static OSErr OpenIPLSConnection(Str255 theIPAddress)
{
	char		justaddress[63];
	char*		cp;
	long		portnum;
	Str31		theport;
	short		index;
	
	//copy over the address and parse out the port
	
	theport[0] = 0;
	cp = justaddress;
	for (index = 1; (index<=theIPAddress[0])&&(theIPAddress[index]!=':'); index++)
	{
		*cp = theIPAddress[index]; cp++;
	}
	
	*cp = 0;	//mark the end of string
	if (index<=theIPAddress[0])	//copy over the port
	{
		index++;	//skip past the colon
		for ( ;index<=theIPAddress[0]; index++)
		{
			if (theport[0]<31)
			{
				theport[0]++; theport[theport[0]] = theIPAddress[index];
			}
		}
	}
	StringToNum(theport, &portnum);
	if (!portnum) portnum = kDefaultIPLSPort;	//our default for now is 1026
	
	
	gTheConnection = new IPLSBROTEndPoint();
	return gTheConnection->Open(gTheIdler, justaddress, portnum);
}

// Each return line has 3 digits, a space, and then the other data/text
static short GetCommandNum(Str255 thestr)
{
	return (thestr[1]-'0')*100 + (thestr[2]-'0')*10 + (thestr[3]-'0');
}

//
// Fill Services.
//
// Sends "SV<cr><lf>" to server.
// Get back multiple lines of "120 tttt Description<cr><lf>" (tttt is the four character type)
// When all services are sent, we get back "190 EOF<cr><lf>"
//
// We just put the result, including the type but not the digits, cr, & lf, into a standard Toolbox list
//

void FillServices(Str255 theIPAddress, ListHandle thelist)
{
	OSErr	theErr;
	Str255	inString;
	char	commandString[256];
	short	commandnum;
	
	Cell	theCell;
	
	theCell.h = 0; theCell.v = 0;
	
	if (OpenIPLSConnection(theIPAddress)==noErr)
	{
		strcpy(commandString, "SV\r\n");	// say to get the list of services
		gTheConnection->SendData(gTheIdler, commandString, strlen(commandString));
		
		do
		{
			theErr = gTheConnection->ReadAString (gTheIdler, inString);
			if (theErr==noErr)
			{
				commandnum = GetCommandNum(inString);
				if (commandnum==120)	//it's an entry
				{
					theCell.h = 0;
					theCell.v=LAddRow(1,(*thelist)->dataBounds.bottom,thelist); // add a row to the bottom
					LSetCell (&inString[5], inString[0]-6, theCell, thelist); //add text removing number & return & linefeed
				}
			}
		}
		while ((theErr==noErr)&&(commandnum!=190));	//loop until 190 file received
		if (theErr==noErr)
		{
			strcpy(commandString, "QU\r\n");
			gTheConnection->SendData(gTheIdler, commandString, strlen(commandString));
			gTheConnection->Close(&gTheIdler);
		}
	}	//of if gTheConnection->open succeeded
	else
	{
		InitCursor();
		StopAlert(131, nil);
	}

	delete gTheConnection; gTheConnection = nil;
}

//
// Fill Names.
//
// Sends "LU<tab>tttt<tab>password<cr><lf>" to server (tttt is the four character type)
// Get back multiple lines of "100 UserName<cr><lf>"
// When all services are sent, we get back "190 EOF<cr><lf>"
//
// We just put the result, including the type but not the cr & lf, into a standard Toolbox list
//
// We don't handle a bad password properly, we just ignore the error.

void FillNames(Str255 theIPAddress, OSType typetofind, Str255 thePassword,  ListHandle thelist, Str31 MyName)
{
	OSErr	theErr;
	Str255	inString;
	char	commandString[256];
	short	commandnum;
	
	Cell	theCell;
	
	theCell.h = 0; theCell.v = 0;
	
	if (OpenIPLSConnection(theIPAddress)==noErr)
	{
		strcpy(commandString, "LU\txxxx\t");	// say to get the list of names
		BlockMove(&typetofind, &commandString[3], 4);	//put in the service to get
		
		thePassword[thePassword[0]+1] = 0;	//make it a nice c string
		strcat(commandString, (char*)&thePassword[1]);	//put in the password
		strcat(commandString, "\r\n");
			
		gTheConnection->SendData(gTheIdler, commandString, strlen(commandString));
		
		do
		{
			theErr = gTheConnection->ReadAString (gTheIdler, inString);
			if (theErr==noErr)
			{
				commandnum = GetCommandNum(inString);
				if (commandnum==100)	//it's an entry
				{
					if (CompareText (&MyName[1], &inString[5], MyName[0], inString[0]-6, nil)!=0)
					{
						theCell.h = 0;
						theCell.v=LAddRow(1,(*thelist)->dataBounds.bottom,thelist); // add a row to the bottom
						LSetCell (&inString[5], inString[0]-6, theCell, thelist); //add text removing number & return & linefeed
					}
				}
				if (commandnum==999)	//Bad password error
				{
					InitCursor();
					StopAlert(129,nil);
				}
			}
		}
		while ((theErr==noErr)&&(commandnum!=190)&&(commandnum<900));	//loop until 190 file received or server complaint
		if (theErr==noErr)
		{
			strcpy(commandString, "QU\r\n");
			gTheConnection->SendData(gTheIdler, commandString, strlen(commandString));
			gTheConnection->Close(&gTheIdler);
		}
	}	//of if gTheConnection->open succeeded
	else
	{
		InitCursor();
		StopAlert(131, nil);
	}

	delete gTheConnection; gTheConnection = nil;
}


//
// RetrieveURL.
//
// Sends "IP<tab>tttt<tab>password<tab>username<cr><lf>" to server (tttt is the four character type)
// 		The protocol supports passwords, I just didn't build it into the browser yet.
// Get back one line of "110 URL<cr><lf>"
//
// We pass the result back in a Str255
//
// We don't handle a bad password properly, we just ignore the error.

void RetrieveURL(Str255 theIPAddress, OSType typetofind, Str255 thePassword,  Str255 thename, Str255 theURL)
{
	OSErr	theErr;
	Str255	inString;
	char	commandString[256];
	short	commandnum;
	
	theURL[0] = 0;
	
	if (OpenIPLSConnection(theIPAddress)==noErr)
	{
		strcpy(commandString, "IP\txxxx\t");	// say to get the URL
		BlockMove(&typetofind, &commandString[3], 4);	//put in the service to get
		
		thePassword[thePassword[0]+1] = 0;	//make it a nice c string
		strcat(commandString, (char*)&thePassword[1]);	//put in the password
		strcat(commandString, "\t");
		
		thename[thename[0]+1] = 0; //turn the name into a C string
		strcat(commandString, (char*)(&thename[1]));	//add it to the command
		strcat(commandString, "\r\n");
		gTheConnection->SendData(gTheIdler, commandString, strlen(commandString));
		
		theErr = gTheConnection->ReadAString (gTheIdler, inString);
		if (theErr==noErr)
		{
			commandnum = GetCommandNum(inString);
			if (commandnum==110)	//it's an entry
			{
				theURL[0] = inString[0] - 6;			//copy the IP over
				BlockMove(&inString[5], &theURL[1], theURL[0]);
			}
			if (commandnum==999)	//Bad password error
			{
				InitCursor();
				StopAlert(129,nil);
			}
		}
		if (theErr==noErr)
		{
			strcpy(commandString, "QU\r\n");
			gTheConnection->SendData(gTheIdler, commandString, strlen(commandString));
			gTheConnection->Close(&gTheIdler);
		}
	}	//of if gTheConnection->open succeeded
	else
	{
		InitCursor();
		StopAlert(131, nil);
	}

	delete gTheConnection; gTheConnection = nil;
}

void LaunchURL(Str255 theURL)
{
	long selstart = 0;
	long selend = theURL[0];
	ICInstance inst;
	ICError theerr;
	
	theerr = ICStart(&inst, 'IPLS');
	if (theerr==noErr)
	{
		theerr = ICFindConfigFile(inst, 0, nil);
		if (theerr==noErr)
			theerr = ICLaunchURL (inst, "\p", (char*)&theURL[1], theURL[0], &selstart, &selend);
		ICStop(inst);
	}
}

// Add a URL to the commandString.  If nil, use our own address.  If Non-nil, use what's there
static void AddTheURL(char* commandString, Str255 theURL, OSType thetype, unsigned short theport)
{
	char	myaddress[32];
	short	index;
	short	count = 0;
	char*	theprotocol;
	Str255	tempstr;
	
	if ((theURL[0]==0)||(theURL[1]==':'))
	{
		switch(thetype)
		{
			case 0:
				//nothing
				break;
			case 'http':
				strcpy(myaddress, "http://");
				count = strlen(myaddress);
				break;
			case 'ftp_':
				strcpy(myaddress, "ftp://");
				count = strlen(myaddress);
				break;
			case 'fing':
				strcpy(myaddress, "finger://");
				count = strlen(myaddress);
				break;
			case 'goph':
				strcpy(myaddress, "gopher://");
				count = strlen(myaddress);
				break;
			default:
				theprotocol = (char*)&thetype;
				for (index = 0; index<4; index++)	//form the protocol into type://URL
				{
					if (theprotocol[index]!='_')
					{
						myaddress[count] = theprotocol[index]; count++;
					}
				}
				myaddress[count] = ':'; count++;
				myaddress[count] = '/'; count++;
				myaddress[count] = '/'; count++;
				myaddress[count] = 0;
				break;
		}
		strcat(commandString, myaddress);
		
		gTheConnection->GetDomainName(myaddress);
		strcat(commandString, myaddress);
		if (theURL[0])
		{
			theURL[theURL[0]+1] = 0;
			strcat(commandString, (char*)&theURL[1]);
		}
	}
	else
	{
		theURL[theURL[0]+1] = 0;
		strcat(commandString, (char*)&theURL[1]);
	}
	if (theport)
	{
		strcat(commandString, ":");
		NumToString(theport, tempstr);
		tempstr[tempstr[0]+1] = 0;
		strcat(commandString, (char*)&tempstr[1]);
	}
}

//
// OK, now for the meat.  We register or unregister each entry in our table, based
// on DoAdd.
//
// When registering, a blank URL will have your machine's address put there, 
// and a URL with a colon has your URL and the url string appended to it
//
// Send a RE<tab>Type<tab>Name<tab>URL<tab>PersPass<tab>RegPass<cr><lf>
// Which at the moment we're treating as
// RE<tab>Type<tab>Name<tab>URL<tab><tab><cr><lf>

OSErr DoRegister(Boolean DoAdd, Str31 theName, Str255 theIPAddress, Str255 thePassword,
	RegisterEntry* thelist, short sizeoflist)
{
	OSErr	theErr = noErr;
	Str255	tempstr;
	char	commandString[256];
	short	commandnum=0;
	short	index;
	OSErr	returnerr=noErr;
	
	if (OpenIPLSConnection(theIPAddress)==noErr)
	{
		for (index = 0; (index<sizeoflist)&&(theErr==noErr)&&(commandnum<900); index++)
		{
			if (thelist[index].myType)
			{
				strcpy(commandString, "RE\tXXXX\t");	// say to register ourselves
				//add the type
				BlockMove(&thelist[index].myType, &commandString[3],4); 
				// add the name
				PLstrcpy(tempstr, theName);
					tempstr[tempstr[0]+1] = 0;
					strcat(commandString, (char*)&tempstr[1]);
				strcat(commandString, "\t");
				if (DoAdd)
				{
					// add the URL
					AddTheURL(commandString, thelist[index].myIPAddress,
						thelist[index].myType, 0);
				}
				strcat(commandString, "\t\t");
				// add the register password
				PLstrcpy(tempstr, thePassword);
					tempstr[tempstr[0]+1] = 0;
					strcat(commandString, (char*)&tempstr[1]);
				strcat(commandString, "\r\n");
				
				gTheConnection->SendData(gTheIdler, commandString, strlen(commandString));
				theErr = gTheConnection->ReadAString (gTheIdler, tempstr);
				if (theErr==noErr)
				{
					commandnum = GetCommandNum(tempstr);
					if (commandnum==999)	//bad password
					{
						InitCursor();
						StopAlert(129, nil);	//tell the user bad password
						returnerr = -2;	//bad password
					}
				}
			}
		}	//for index
		if (theErr==noErr)
		{
			strcpy(commandString, "QU\r\n");
			gTheConnection->SendData(gTheIdler, commandString, strlen(commandString));
			gTheConnection->Close(&gTheIdler);
//			if (commandnum<900)
//				WeAreRegistered = DoAdd;	//remember that we've registered ourselves
		}
		else returnerr=theErr;
	}	//of if gTheConnection->open succeeded
	else
	{
		InitCursor();
		StopAlert(131, nil);
		returnerr = -1;	//unable to open
	}

	delete gTheConnection; gTheConnection = nil;
	return returnerr;
}

OSErr DoQTLKRegister(Boolean DoAdd, Str31 theName, Str255 theIPAddress, unsigned short theport, Str255 thePassword, OSType thetype)
{
	OSErr	theErr = noErr;
	Str255	tempstr;
	char	commandString[256];
	short	commandnum=0;
	OSErr	returnerr=noErr;
	
	if (OpenIPLSConnection(theIPAddress)==noErr)
	{
		strcpy(commandString, "RE\tXXXX\t");	// say to register ourselves
		//add the type
		BlockMove(&thetype, &commandString[3],4); 
		// add the name
		PLstrcpy(tempstr, theName);
			tempstr[tempstr[0]+1] = 0;
			strcat(commandString, (char*)&tempstr[1]);
		strcat(commandString, "\t");
		if (DoAdd)
		{
			// add the URL
			AddTheURL(commandString, "\p", thetype, theport);
		}
		strcat(commandString, "\t\t");
		// add the register password
		PLstrcpy(tempstr, thePassword);
			tempstr[tempstr[0]+1] = 0;
			strcat(commandString, (char*)&tempstr[1]);
		strcat(commandString, "\r\n");
		
		gTheConnection->SendData(gTheIdler, commandString, strlen(commandString));
		theErr = gTheConnection->ReadAString (gTheIdler, tempstr);
		if (theErr==noErr)
		{
			commandnum = GetCommandNum(tempstr);
			if (commandnum==999)	//bad password
			{
				InitCursor();
				StopAlert(129, nil);	//tell the user bad password
				returnerr = -2;	//bad password
			}
		}
		if (theErr==noErr)
		{
			strcpy(commandString, "QU\r\n");
			gTheConnection->SendData(gTheIdler, commandString, strlen(commandString));
			gTheConnection->Close(&gTheIdler);
		}
		else returnerr=theErr;
	}	//of if gTheConnection->open succeeded
	else
	{
		InitCursor();
		StopAlert(131, nil);
		returnerr = -1;	//unable to open
	}

	delete gTheConnection; gTheConnection = nil;
	return returnerr;
}

