/**********************************************************************************
	FILE: OTHandlerEndPoint.cp
	Description: Handlers receive connections from the server, read text from the
		connection and send it to HandleData, and handle sending text down the line

	1997, CE Software, Inc.  All rights reserved.
	P.O. Box 65580
	West Des Moines, Iowa 50265 U.S.A.

***********************************************************************************/

#include "OTServerEndPoint.h"
#include "OTHandlerEndPoint.h"

// Our constructor opens and endpoint and tells the server I'm ready
OTHandlerEndPoint::OTHandlerEndPoint (OTServerEndPoint* theserver, short handlerID)
{
	OSErr		err = noErr;
			
	mfCancelled = false;
	mfDataWaiting = false;
	mfNeedsDisconnecting = false;
	mfJustConnected = false;
	mpEndpoint = nil;
	mfNeedsUnbinding = false;
	mfDataInIsText = true;
	
	mfServer = theserver;
	mfHandlerID = handlerID;
	
	mfOutGoingData = nil;	//prepare for sending
	mfSizeOfData = 0;
	mfBytesSent = 0;
	
	mfBytesReceived = 0;
	
	meState = keOpeningListener;
	
	OTAsyncOpenEndpoint( OTCreateConfiguration(kTCPName) ,0, &mfEndpointInfo,&Notifier, this);
	
	//Allocate enough info for our binding info
	size_t buffSize	= mfEndpointInfo.addr 
										+ mfEndpointInfo.options 
					 					+ (mfEndpointInfo.connect > mfEndpointInfo.discon 
					 								? mfEndpointInfo.connect
					 								: mfEndpointInfo.discon);

	mfData = new UInt8[buffSize];
}

// Destructor
OTHandlerEndPoint::~OTHandlerEndPoint (void)
{
	DestroyEndPoint();
	
	if( mfData) delete mfData;
}

#pragma push
#pragma segment Main

//
// When the server has a connection for us to handle, it calls our PassOff method
//
void OTHandlerEndPoint::PassOff(EndpointRef theref)
{
	meState = keAccepting;
	mfCancelled = false;
	mfDataWaiting = false;
	mfNeedsDisconnecting = false;
	mfJustConnected = false;
	mfNeedsUnbinding = false;
	
	mfCallInfo.addr.maxlen = mfEndpointInfo.addr;
	mfCallInfo.addr.len = 0;
	mfCallInfo.opt.maxlen = mfEndpointInfo.options;
	mfCallInfo.opt.buf = nil;
	mfCallInfo.udata.maxlen 	= mfEndpointInfo.connect;
	mfCallInfo.udata.len = 0;
	mfCallInfo.udata.buf = nil;
	
	mfDisconInfo.udata.maxlen		= mfEndpointInfo.discon;
	mfDisconInfo.udata.len		= 0;
	
// Alloacte enough ram for Tcall & TDiscon buffers
	size_t buffSize	= mfCallInfo.addr.maxlen 
										+ mfCallInfo.opt.maxlen 
					 					+ (mfCallInfo.udata.maxlen > mfDisconInfo.udata.maxlen 
					 								? mfCallInfo.udata.maxlen
					 								: mfDisconInfo.udata.maxlen);

	if (mfData)
		mfData = new UInt8[buffSize];

	mfCallInfo.addr.buf 	 		= &mfData[0];
	mfCallInfo.opt.buf 	 		= mfCallInfo.addr.buf + mfCallInfo.addr.maxlen;
	mfCallInfo.udata.buf 	 	= mfCallInfo.opt.buf + mfCallInfo.opt.maxlen;
	mfDisconInfo.udata.buf 	= mfCallInfo.udata.buf; 	// Disconnect and connect share the same data area

	OTListen(theref, &mfCallInfo);				//OK, listen and accept it
	OTAccept(theref, mpEndpoint, &mfCallInfo);

}

// Read and send data

void OTHandlerEndPoint::DoIdle(void)
{
	OSStatus	theErr;
	
	if ((mfNeedsUnbinding)&&( EndPointValid() ))
	{
		theErr = OTUnbind(mpEndpoint);
		if (theErr!=kOTStateChangeErr)
		{
			mfNeedsUnbinding = false;
			HandleShutdown();
		}
	}
	
	if (mfJustConnected)
	{
		mfJustConnected = false;
		mfDataInIsText = true;		//always start off a connection in text format
		HandleStartup();
	}
	
	//If we're in the middle of a send, send it
	SendMore();
	
	//Check and see if we've received a string
	GetData();
}

// Returns true if the ~OTHandlerEndPoint::Cancel() has been called since the endpoint was opened, OR
// if the connection has been broken (by the other end disconnecting, for example)
// In keeping with the Connection class cancelling is an immediate, synchronous operation. Cancelling
// is expected to cause all further operations to fail immediately, thus eventually exiting out of the
// operation.
Boolean OTHandlerEndPoint::EndPointValid (void) const
{
	return (mpEndpoint != nil && !mfCancelled);
}


OSErr OTHandlerEndPoint::DestroyEndPoint (void)
{
	OSErr		iErr = noErr;
	
	if ((mpEndpoint !=  kOTInvalidEndpointRef)&&(mpEndpoint!=nil))
	{
		iErr = OTCloseProvider(mpEndpoint);
		mpEndpoint = nil;
		mfDataWaiting = false;
	}
	return iErr;
}

void OTHandlerEndPoint::CheckOTError (OTResult lError)
{
	if (lError == kOTLookErr)
	{
		if (mpEndpoint != nil)
		{
			OTResult		lResult;
			
			switch (OTLook (mpEndpoint))
			{
				case T_ORDREL:
					lResult = OTRcvOrderlyDisconnect (mpEndpoint);
					break;
				case T_DISCONNECT:
					lResult = OTRcvDisconnect (mpEndpoint, nil);
					break;
			}
		}
	}
}
#pragma pop


// Put call back in main segment to avoid having the segment unloaded and causing typical callback problems
// if the segment is unloaded when the callback is called
#pragma push
#pragma segment Main

// *** Called at deferred task time. ***
pascal void OTHandlerEndPoint::Notifier(void* contextPtr, OTEventCode code, OTResult result, void* cookie)
{
	OTHandlerEndPoint*	thisPoint = (OTHandlerEndPoint*) contextPtr;
	thisPoint->HandleNotify (code, result, cookie);
}


// *** Called at deferred task time. ***
void OTHandlerEndPoint::HandleNotify( OTEventCode code, OTResult result, void* cookie)
{
	switch (code)
	{
		case T_OPENCOMPLETE:
			miAsyncErr = result;
			mpEndpoint = (EndpointRef)cookie;
			meState = keOpenListenerDone;
			mfServer->SetHandlerStatus(mfHandlerID, true);	//OK, we're ready to be connected to
			break;
		case T_PASSCON:	//we've started our connection
			miAsyncErr = result;
			meState = keAcceptingDone;
			mfJustConnected = true;
			break;
		// the other side broke the connection, so disconnect us but don't delete us
		case T_DISCONNECT:
			miAsyncErr = result;
			meState = keDisconnecting;
			// Disconnect us, but don't delete us
			OTRcvDisconnect(mpEndpoint, &mfDisconInfo); 
			// prepare to unbind when it's safe to allocate memory
			mfNeedsUnbinding = true;
			meState = keDisconnectingDone;
			break;
		// the other side is requestion a break of connection (how polite).
		case T_ORDREL:
			miAsyncErr = result;
			meState = keOrderlyDisconnecting;
			OTRcvOrderlyDisconnect (mpEndpoint);
			if (OTGetEndpointState(mpEndpoint)!=T_IDLE)
			{
				OTSndOrderlyDisconnect(mpEndpoint);
			}
			// prepare to unbind when it's safe to allocate memory
			mfNeedsUnbinding = true;
			meState = keOrderlyDisconnectingDone;
			break;
		case T_DISCONNECTCOMPLETE:
			// prepare to unbind when it's safe to allocate memory
			mfNeedsUnbinding = true;
			meState = keBinding;
			break;
		case T_UNBINDCOMPLETE:	//OK, we're set for another connection
			mfServer->SetHandlerStatus(mfHandlerID, true);
			meState = keOpenListenerDone;
			mfNeedsUnbinding = false;
		case T_DATA:
			mfDataWaiting = true;
			break;			
	}
}
//
// If we need to send some MOD or other text on connection, override this
//
void OTHandlerEndPoint::HandleStartup(void)
{
	
}
//
// If we need to tell someone we've closed, override this
//
void OTHandlerEndPoint::HandleShutdown	(void)
{

}

#pragma pop

// Note that Close does NOT kill the endpoint, because we need to be recycled into the server
OSErr OTHandlerEndPoint::Close (void)
{
	OSErr		iResult = noErr;
	
	OTResult state = ::OTGetEndpointState(mpEndpoint);
	
	if(state == T_DATAXFER)
	{
		// disconnect the endpoint
		if( (mfEndpointInfo.flags == T_COTS_ORD) || (mfEndpointInfo.flags == T_TRANS_ORD))
			::OTSndOrderlyDisconnect(mpEndpoint);
		else 
			::OTSndDisconnect (mpEndpoint,&mfCallInfo);
	}

	return iResult;
}


void OTHandlerEndPoint::Cancel (void)
{
	mfCancelled = true;
}

//
// See if we've got our SendData port open.  If so, copy the data and start sending it
//
OSErr OTHandlerEndPoint::SendData (const void *pData, long lSize)
{
	long		lBytesSent = 0;
	OSStatus	lResult = noErr;
	
	if (mfOutGoingData)	return -1;	//still sending the last stuff
	if (!EndPointValid()) return -1;
	
	mfSizeOfData = lSize;
	mfBytesSent = 0;
	mfOutGoingData = NewPtr(mfSizeOfData);
	if (mfOutGoingData)
	{
		BlockMove(pData, mfOutGoingData, mfSizeOfData);
	}
	
	return SendMore();
}
//
// If the data is too big to be sent at once, this sends it in chunks.
// Called by DoIdle (and SendData for the first chunk)
//
OSErr OTHandlerEndPoint::SendMore (void)
{
	long		lBytesSent = 0;
	OSStatus	lResult = noErr;
	
	if (!mfOutGoingData) return noErr;
	if (!EndPointValid()) return -1;
	
	lResult = OTSnd (mpEndpoint, (void*) (((UInt8*) mfOutGoingData) + mfBytesSent),
		 mfSizeOfData-mfBytesSent, 0);
		
	if (lResult >= 0)
	{
		mfBytesSent += lResult;
		if (mfBytesSent >= mfSizeOfData)
		{
			DisposePtr(mfOutGoingData);
			mfOutGoingData = nil;
		}
		return noErr;
	}

	return (OSErr) lResult;
}
//
// This is the workhorse, it gets the commands and responds to it, when you
// override it!
//
void OTHandlerEndPoint::HandleData (char* /*thestring*/, short /*numchars*/)
{
	//do nothing here, only for handled stuff
}
//
// Read data and pass it off to HandleData, called by DoIdle()
//
OSErr OTHandlerEndPoint::GetData ( void )
{

	OSStatus	lResult = noErr;
	OTFlags		flags;
	OTResult	theResult;
	
	short	index;
	
	if (!EndPointValid ())
		return -1;
	if ((!mfDataWaiting)&&(!mfBytesReceived))
		return noErr;
	
	theResult = OTRcv (mpEndpoint, (void*) (&mfInComingData[mfBytesReceived]),
		 sizeof(mfInComingData) - mfBytesReceived, &flags);
	
	if (theResult == 0)
	{
		mfDataWaiting = false;
		lResult = noErr;
	}
	if (theResult >= 0)
	{
		mfBytesReceived += theResult;
		
		if ((flags & T_MORE) == 0)
		{
			mfDataWaiting = false;
		}
		
		if ((sizeof(mfInComingData)==mfBytesReceived)||(!mfDataInIsText))	//we filled our buffer or we don't parse
		{
			this->HandleData(mfInComingData, mfBytesReceived);	//have them handle it and clear it
			mfBytesReceived = 0;
		}
		else
		{
			for (index = 0; index < mfBytesReceived; index++)
			{
				if (mfInComingData[index]==0x0A)	//found end of line, line feed of cr/lf
				{
					index++;	//convert the index into the size
					this->HandleData(mfInComingData, index);	//have them handle it and clear it
					BlockMove(&mfInComingData[index], &mfInComingData[0], mfBytesReceived-index);
					mfBytesReceived-=index;
					index = 0;
				}
			}
		}
		
		lResult = noErr;
	}
	else if (theResult != kOTNoDataErr)
	{
		CheckOTError (theResult);
		lResult = theResult;
	}
	else
	{
		// else handle kOTNoDataErr
		lResult = noErr;
		mfDataWaiting = false;
	}

	return (OSErr) lResult;
}

