/**********************************************************************************
	FILE: OTServerEndPoint.cp
	Description: This server exists to register a port and listen for TCP/IP connection.
		It then hands those requests to handler classes

	1997, CE Software, Inc.  All rights reserved.
	P.O. Box 65580
	West Des Moines, Iowa 50265 U.S.A.

***********************************************************************************/

#include "OTServerEndPoint.h"
#include "OTHandlerEndPoint.h"

//
// Our Constructor.  Starts asynchronously opening the enpoint, and zeros our array
// of handlers (just in case they somehow get called before they get started.
//
OTServerEndPoint::OTServerEndPoint (unsigned short port)
{
	short		index;
			
	mfPort = port;	//save our port off
	
	mfCancelled = false;
	mfNeedsUnBinding = false;
	mpEndpoint = nil;
	
	meState = keOpeningListener;
	
	OTAsyncOpenEndpoint( OTCreateConfiguration(kTCPName) ,0, nil,&Notifier, this);
	
	// We create our handlers later. 
	for (index = 0; index<kMaxHandlers; index++)
	{
		mfhandler[index].Unused = false; //Set to true when the endpoint is fully open
		mfhandler[index].h = nil;
	}
}
//
// Create our handlers, by calling a method that can be overridden.
//
// This cannot be part of the constructor, because the overridden method
// would not be called from our constructor
//
void OTServerEndPoint::StartHandlers(void)
{
	short	index;
	
	// Create our handlers.   We call a method to create them so it can be overridden
	for (index = 0; index<kMaxHandlers; index++)
	{
		mfhandler[index].Unused = false; //Set to true when the endpoint is fully open
		mfhandler[index].h = this->CreateHandler(index);
	}
}

// Create a handler.  This routine should always be overridden by a subclass,
// to create the actual handlers.
OTHandlerEndPoint* OTServerEndPoint::CreateHandler (short HandlerID)
{
	return new OTHandlerEndPoint(this, HandlerID);
}

// Destructor
OTServerEndPoint::~OTServerEndPoint (void)
{
	short	index;
	
	// Dispose our handlers.
	for (index = 0; index<kMaxHandlers; index++)
	{
		delete mfhandler[index].h;
		mfhandler[index].h = nil;
	}
	
	DestroyEndPoint();	//Unbind us, and close the endpoint
}

// Idle routine.  It calls our Handlers' idle routines, which read and write
// data to the Internet.
void OTServerEndPoint::DoIdle (void)
{
	short		index;
	
	OTIdle();
			
	for (index = 0; index<kMaxHandlers; index++)
	{
		if (mfhandler[index].h)
		{
			mfhandler[index].h->DoIdle();
		}
	}
}

#pragma push
#pragma segment Main

OSErr OTServerEndPoint::DestroyEndPoint (void)
{
	OSErr		iErr = noErr;
	
	if ((mpEndpoint !=  kOTInvalidEndpointRef)&&(mpEndpoint!=nil))
	{
		if (mfNeedsUnBinding)
		{
			meState = keUnBinding;
			if (OTUnbind(mpEndpoint)==noErr)
				do {} while (meState == keUnBinding);
		}
		iErr = OTCloseProvider(mpEndpoint);
		mpEndpoint = nil;
	}
	return iErr;
}

void OTServerEndPoint::CheckOTError (OTResult lError)	//disconnect if necessary
{
	if (lError == kOTLookErr)
	{
		if (mpEndpoint != nil)
		{
			OSErr			iDummy;
			OTResult		lResult;
			
			switch (OTLook (mpEndpoint))
			{
				case T_ORDREL:
					lResult = OTRcvOrderlyDisconnect (mpEndpoint);
					iDummy = DestroyEndPoint ();
					break;
				case T_DISCONNECT:
					lResult = OTRcvDisconnect (mpEndpoint, nil);
					iDummy = DestroyEndPoint ();
					break;
			}
		}
	}
}
#pragma pop


// Put call back in main segment to avoid having the segment unloaded and causing typical callback problems
// if the segment is unloaded when the callback is called
#pragma push
#pragma segment Main

// *** Called at deferred task time. ***
pascal void OTServerEndPoint::Notifier(void* contextPtr, OTEventCode code, OTResult result, void* cookie)
{
	OTServerEndPoint*	thisPoint = (OTServerEndPoint*) contextPtr;
	thisPoint->HandleNotify (code, result, cookie);
}


// *** Called at deferred task time. ***
void OTServerEndPoint::HandleNotify( OTEventCode code, OTResult result, void* cookie)
{
	switch (code)
	{
		case T_OPENCOMPLETE:	//open finished. Bind us to our port 
			miAsyncErr = result;
			meState = keOpenListenerDone;
			HandleOpenComplete(cookie);
			break;
		case T_BINDCOMPLETE:	//Binding done, we now just listen
			miAsyncErr = result;
			meState = keListening;
			HandleBindComplete();
			break;
		case	T_LISTEN:			//Someone wants to talk to us
			miAsyncErr = result;
			meState = keAccepting;
			HandleListen();
			break;
		case	T_ACCEPTCOMPLETE:	//We have finished handing the connection off
			miAsyncErr = result;
			meState = keListening;
			break;
		case T_UNBINDCOMPLETE:	//closing down, we're ready to be closed
			meState = keUnBound;
			mfNeedsUnBinding = false;
			break;
		
	}
}

//  The open is done, bind us to the port
void OTServerEndPoint::HandleOpenComplete( void* cookie )
{
	
	TBind		inBind = { { sizeof(InetAddress),
					 sizeof(InetAddress),
					 (UInt8*) &mfInAddress }, 1 };
	TBind		outBind = { { sizeof(InetAddress),
					 sizeof(InetAddress),
					 (UInt8*) &mfOutAddress }, 1 };
	OSStatus	err;
	
	mpEndpoint = (EndpointRef)cookie;
	
	inBind.qlen = kMaxHandlers;
	
	OTInitInetAddress(&mfInAddress, mfPort, kOTAnyInetAddress);
	OTInitInetAddress(&mfOutAddress, mfPort, kOTAnyInetAddress);
	err = OTBind (mpEndpoint, &inBind, &outBind);
}

// Remember the port we actually got (if we passed port 0) and mark that
// we must be UnBinded sometime
void OTServerEndPoint::HandleBindComplete( void)
{
	mfPort = mfOutAddress.fPort;
	mfNeedsUnBinding = true;
}

//  Got a call.  Find an open handler and hand it off
void OTServerEndPoint::HandleListen( void)
{
	
	OTHandlerEndPoint* thehandler = FindOpenHandler();
	
	if (thehandler)
	{
		thehandler->PassOff(mpEndpoint);
	}

}

#pragma pop

OSErr OTServerEndPoint::Close (void)
{
	OSErr		iResult = noErr;
	
	// Need to close the endpoint if Cancel() was called so we check for mpEndpoint != nil instead of EndPointValid ()
	if (mpEndpoint != nil)
	{
		OSErr		iDummy;
		
		iDummy = DestroyEndPoint ();
	}
	return iResult;
}


void OTServerEndPoint::Cancel (void)
{
	mfCancelled = true;
}

// Handlers call this to tell the server that they are ready to accept a
// communication.  Called first when they have finished opening, then called
// when they have disconnected from a communication
void OTServerEndPoint::SetHandlerStatus (short HandlerID, Boolean IsUnused)
{
	mfhandler[HandlerID].Unused = IsUnused;
}

// Find an open handler for a connection
OTHandlerEndPoint* OTServerEndPoint::FindOpenHandler (void)
{
	short	index;
	
	for (index = 0; index<kMaxHandlers; index++)
	{
		if (mfhandler[index].Unused) //Is this open?
		{
			mfhandler[index].Unused = false;	//yup, mark we've used it
			return mfhandler[index].h;			//and pass it back
		}
	}
	return nil;
}

