#ifndef OTSERVERENDPOINT_H
#define OTSERVERENDPOINT_H

/**********************************************************************************
	FILE: OTServerEndPoint.h
	Description: Generic server endpoint class.  This server exists to register
		a port and listen for TCP/IP connection.  It then hands those requests
		to handler classes

	1997, CE Software, Inc.  All rights reserved.
	P.O. Box 65580
	West Des Moines, Iowa 50265 U.S.A.
	
	DESIGN NOTES:  Most clients want to execute transactions synchronously--issue
		a command, get the result, idling all the while to let other processes do
		things.  But, servers need to be more asynchronous, particularly to support
		multiple concurrent connections.
		
		This server class and the partner OTHandlerEndPoint class work by the
		server listening for connection requests and passing them off to a handler.
		The connection takes place through the OT Notifier routine.
		
		Reading and writing of data is done through an Idle routine.  The server's
		idle routine calls the handlers idle routines.  The Handler idle routines
		send data if any is waiting to go out, and then look for incoming data.
		The incoming data is parsed into separate lines, and passed to a method
		HandleData.  (This won't work for non-textual data transfers.)
	
	IMPLEMENTATION NOTES:  To use these classes, you create a derived class from
		OTServerEndPoint where you override the CreateHandler method.  Your version
		of CreateHandler just creates your derived subclass of OTHandlerEndPoint.
		
		Your derived class from OTHandlerEndPoint overrides the HandleData method.
		It is periodically called with a string that it is to respond to.
		
		Your main application simply creates a server object (passing it the port
		number to use), calls the Idle method in your main event loop, and dispose
		of the object when you quit.

***********************************************************************************/

#include <OpenTransport.h>
#include <OpenTptInternet.h>
#include <Types.h>

// This defines how many handlers and thus how many concurrent connections we can handle.
// Someday this may become a parameter with a dynamic array
#define kMaxHandlers 10

// The handlers that do the actual work
class OTHandlerEndPoint;

struct HandlerRec {
	
	OTHandlerEndPoint*	h;			//this handles it all
	Boolean				Unused;		//this says whether this is ready to accept a new version
};
typedef struct HandlerRec HandlerRec;

class OTServerEndPoint
{
	
	public:
		//+++Enums

	private:
		Boolean			mfCancelled;
		Boolean			mfDataWaiting;
		Boolean			mfNeedsUnBinding;
		volatile EndpointRef		mpEndpoint;
		
		unsigned short mfPort;
		InetAddress	mfInAddress;
		InetAddress	mfOutAddress;

		HandlerRec		mfhandler[kMaxHandlers];
		
		
		enum	OutpointState	{	keBinding, keBindingDone,
										keConnecting, keConnectingDone,
										keDisconnecting, keDisconnectingDone,
										keOrderlyDisconnecting, keOrderlyDisconnectingDone,
										
										keOpeningMapper, keOpeningMapperDone,
										keResolving, keResolvingDone,
										
										keOpeningListener, keOpenListenerDone,
										keListening, keAccepting, keAcceptingDone,
										
										keBailingOut, keUnBinding, keUnBound	};
		
		OSErr				miAsyncErr;
		OutpointState	meState;
		union {
			InetAddress	inet;
		}					mfDestAddress;
 	
	protected:
						OTServerEndPoint		(const OTServerEndPoint& oRHS);
		OTServerEndPoint&	operator=		(const OTServerEndPoint& oRHS);
		
		OSErr			DestroyEndPoint(void);
		void			CheckOTError	(OTResult lError);
		
		virtual OTHandlerEndPoint*	CreateHandler(short HandlerID);
		OTHandlerEndPoint*	FindOpenHandler (void);
		
		static pascal void		Notifier			(void* contextPtr, OTEventCode code,
															OTResult result, void* cookie);
		void		 	HandleNotify	(OTEventCode code, OTResult result, void* cookie);
		void			HandleOpenComplete(void* cookie);
		void			HandleBindComplete(void);
		void			HandleListen(void);

	public:
						OTServerEndPoint		(unsigned short port);
		virtual			~OTServerEndPoint		(void);
		
		void			StartHandlers(void);
		
		OSErr			Close				(void);

		void			Cancel			(void);
		void			SetCancelFlag 	(Boolean fCancel) { mfCancelled = fCancel; }
		void			DoIdle			(void);
		
		void			SetHandlerStatus(short HandlerID, Boolean IsUnused);
};

#endif
