/*
	File:		SubwooferProgress.cpp

	Contains:	<contents>

	Written by:	Frederic Artru

		Copyright:	 1997 by Apple Computer, Inc., all rights reserved.

     You may incorporate this sample code into your applications without
    restriction, though the sample code has been provided "AS IS" and the
    responsibility for its operation is 100% yours.  However, what you are
    not permitted to do is to redistribute the source as "DSC Sample Code"
    after having made changes. If you're going to re-distribute the source,
    we require that you make it clear in the source that the code was
    descended from Apple Sample Code, but that you've made changes.

	Change History (most recent first):

         <7>     3/12/97    FA      nil progress struct in DisposeProgress
         <6>     3/11/97    FA      fix to status string
         <5>      3/9/97    FA      No downloading string appears when amountDone is null
         <4>      3/7/97    FA      would erase status string when it didn't change
         <3>      3/5/97    FA      if no status string or amountdone, don't change status field.
                                    Added CompareStrings to clear the code a little.
         <2>      3/4/97    FA      Making it localizable. Also adding the download file name and
                                    percent download.
         <1>      3/3/97    FA      first checked in
*/


#include <QuickDraw.h>
#include <TextUtils.h>
#include <Resources.h>
#include <Files.h>
#include <ToolUtils.h>
#include <OSUtils.h>
#include <Scrap.h>
#include <fp.h>

#include "Subwoofer.h"
#include "SubwooferProgress.h"

#include "SWUtilities.h"

// prototypes
void GetAmountString(UInt32 amount, Str255 theString);
void DrawProgressBar(ProgressDataStruct *data, UInt32 amountDone, UInt32 amountTotal);
void AppendString(StringPtr dest, StringPtr source);
Boolean CompareStrings(StringPtr a, StringPtr b);

OSErr CreateProgress(ProgressDataStruct* progressData, Boolean displayProgess, StringPtr fileName)
{
	OSErr err = noErr;
	if (!displayProgess)
	{
		// The progress dialog will not be displayed...
		// ... clear the data
		progressData->progressDialog = nil;
		UInt16 index = kMinBarberPolePictID;
		while ( index <= kMaxBarberPolePictID )
		{
			progressData->barberPolePict[index - kMinBarberPolePictID] = nil;
			index++;
		}
		
		progressData->whichPict = 0;
	}
	else
	{
		// open Subwoofer's resource fork (REPLACE FOR SAMPLE CODE)
		err = OpenSubwooferForResource();
		
		if ( err == noErr )
		{		
			// load the dialog
			progressData->progressDialog = GetNewDialog(kProgressDialogResID, nil, (WindowPtr) -1);
			
			// Set the file name field
			Handle theItem;
			GetDialogItem(progressData->progressDialog, kProgressFileNameItem, nil, &theItem, nil);
			SetDialogItemText(theItem, fileName);
			
			// Set the cancel button
			SetDialogCancelItem(progressData->progressDialog, kProgressCancelButtonItem);
			
			// Now show the dialog
			ShowWindow(progressData->progressDialog);
			
			// get the barber pole pictures
			UInt16 index = kMinBarberPolePictID;
			while ( index <= kMaxBarberPolePictID )
			{
				progressData->barberPolePict[index - kMinBarberPolePictID] = GetPicture(index);
				index++;
			}
			
			progressData->whichPict = 0;
		}
		
		// close Subwoofer's resource fork (REPLACE FOR SAMPLE CODE)
		CloseSubwooferForResource();
	}
	return err;
}


void DisposeProgress(ProgressDataStruct* progressData)
{
	if (progressData->progressDialog != nil)
	{
		DisposeDialog(progressData->progressDialog);
		
		UInt16 index = kMinBarberPolePictID;
		while ( index != kMaxBarberPolePictID )
		{
			ReleaseResource((Handle) progressData->barberPolePict[index - kMinBarberPolePictID]);
			index++;
		}
		
		progressData->progressDialog = nil;
	}
}

Boolean DoProgress(ProgressDataStruct* progressData, UInt32 amountDone, UInt32 amountTotal, StringPtr statusString)
{
	Boolean wasCancelled = false;

	if (progressData->progressDialog != nil)
	{
		Str255 downloadString;
		Str255 tempString;
		GrafPtr savePort;

		GetPort(&savePort);
		SetPort(progressData->progressDialog);
		
		// calculate download string if we can (e.g.: "867k of 1.1Mb (77%)")
		downloadString[0] = 0;
		AppendString(downloadString, "\pDownloading...");
		if ( amountDone > 0 )
		{
			// append done amount
			GetAmountString(amountDone, tempString);
			AppendString(downloadString, "\p ");
			AppendString(downloadString, tempString);
			
			if ( amountTotal > 0 )
			{
				// append total amount
				GetAmountString(amountTotal, tempString);
				AppendString(downloadString, "\p of ");
				AppendString(downloadString, tempString);
				
				// append percent
				NumToString(100 * amountDone / amountTotal, tempString);
				AppendString(downloadString, "\p (");
				AppendString(downloadString, tempString);
				AppendString(downloadString, "\p%)");
			}
		}
		
		// set the download items
		// if the status is Downloading, we append the downloading string.
		// otherwise, only display the status string
		Handle theItem;
		GetDialogItem(progressData->progressDialog, kProgressDownloadingItem, nil, &theItem, nil);
		GetDialogItemText(theItem, tempString);
		
		Boolean statusIsDownloading = CompareStrings(statusString, "\pDownloading...");
		Boolean downloadingStringHasChanged = !CompareStrings(tempString, downloadString);
		
		if ( statusString[0] && !statusIsDownloading )
			SetDialogItemText(theItem, statusString);
		else if ( downloadingStringHasChanged && amountDone > 0 ) 
			SetDialogItemText(theItem, downloadString);
						
		// now do the progress bar
		DrawProgressBar(progressData, amountDone, amountTotal);
		
		EventRecord event;
		if ( WaitNextEvent(everyEvent, &event, 0, nil) )
		{
			
			if (IsDialogEvent(&event))
			{
				SInt16 itemHit;
				
				// Ask the standard filter proc to handle keyboard events (command-period, etc...)
				if ( StdFilterProc(progressData->progressDialog, &event, &itemHit) )
				{
					wasCancelled = (itemHit == kProgressCancelButtonItem);
				}
				else
				{
					DialogPtr theDialog;
					if ( DialogSelect(&event, &theDialog, &itemHit) )
					{
						wasCancelled = (theDialog == progressData->progressDialog && 
										itemHit == kProgressCancelButtonItem);
					}
				}
			}
		}

		SetPort(savePort);
	}
	
	return wasCancelled;
}

void GetAmountString(UInt32 amount, Str255 theString)
{
	// All the strings are saved in a Subwoofer resource (REPLACE THEM FOR SAMPLE CODE)
	UInt32 temp;
	Str32 unit = "\p";
	
	// see if we have bytes, kilobytes, megabytes or gigabytes (yeah, right !)
	if ( amount == 0 )
	{
		BlockMoveData("\p0 bytes", theString, 8);
	}
	else
	{
		if ( amount < 1024 )
		{
			temp = amount;
			if ( amount == 1 )
				AppendString(unit, "\p byte");
			else
				AppendString(unit, "\p bytes");
		}
		else if ( amount < 1024*1024 )
		{
			temp = amount / 1024;
			AppendString(unit, "\p kbytes");
		}
		else
		{
			temp = amount / 1024 / 1024;
			UInt32 reste = amount % (1024 * 1024) / (1024 * 1024 /10); // get the 100Ks
			Str32 resteStr;
			NumToString(reste, resteStr);
			AppendString(unit, "\p.");
			AppendString(unit, resteStr);
			AppendString(unit, "\p Mbytes");
		}
		
		NumToString(temp, theString);
		AppendString(theString, unit);
	}
}

void DrawProgressBar(ProgressDataStruct *data, UInt32 amountDone, UInt32 amountTotal)
{
	SInt16			theType;
	UInt16			theLength;
	float			floatDone, floatMax;
	Handle			theHan;
	Rect			progressBarRect;

	
	GetDialogItem ( data->progressDialog, kProgressBarItem, &theType, &theHan, &progressBarRect );
	if ( amountTotal > 0 )		// Standard Progress Bar
	{		
		theLength = progressBarRect.right - progressBarRect.left;
		
		floatDone = amountDone;
		floatMax = amountTotal;
		
		// Draw the rectangle of what's done
		Rect doneRect = progressBarRect;
		doneRect.right = doneRect.left + rinttol((floatDone / floatMax) * theLength);
		RGBColor doneColor = { 0x4444, 0x4444, 0x4444 };
		RGBForeColor(&doneColor);
		PaintRect(&doneRect);
		
		// Draw the rectangle of what's left to do
		Rect notDoneRect = progressBarRect;
		notDoneRect.left = doneRect.right;
		
		RGBColor backgroundColor = { 0xCEEE, 0xCEEE, 0xFFFF };
		RGBForeColor(&backgroundColor);
		PaintRect(&notDoneRect);
		
		// Reset color to black
		ForeColor( blackColor );
		
		// Draw a frame
		FrameRect(&progressBarRect);
	}
	else					// Barber Pole Progress Bar
	{
		PicHandle		thePic;
		
		thePic = data->barberPolePict[data->whichPict];
		DrawPicture ( thePic, &progressBarRect );
		data->whichPict++;
		if ( data->whichPict > kMaxBarberPolePictID - kMinBarberPolePictID )
			data->whichPict = 0;
	}
}

void AppendString(StringPtr dest, StringPtr source)
{
	short sourceLen = source[0];
	short destLen = dest[0];
	
	BlockMoveData(&(source[1]), &(dest[destLen + 1]), sourceLen);
	dest[0] = destLen + sourceLen;
}

Boolean CompareStrings(StringPtr a, StringPtr b)
{
	Boolean result = (a[0] == b[0]);
	
	if ( result )
	{
		UInt16 index = a[0];
		while ( index && result )
		{
			result = result && ( a[index] == b[index] );
			index--;
		}
	}
	
	return result;
}
