/*
	File:		SubwooferURLDownload.cpp

	Contains:	<contents>

	Written by:	Frederic Artru

		Copyright:	 1997 by Apple Computer, Inc., all rights reserved.

     You may incorporate this sample code into your applications without
    restriction, though the sample code has been provided "AS IS" and the
    responsibility for its operation is 100% yours.  However, what you are
    not permitted to do is to redistribute the source as "DSC Sample Code"
    after having made changes. If you're going to re-distribute the source,
    we require that you make it clear in the source that the code was
    descended from Apple Sample Code, but that you've made changes.

	Change History (most recent first):

        <16>      4/7/97    FA      read long in SaveBufferToFile. Create URLReference before
                                    creating file.
        <15>     3/12/97    FA      fixing with file spec
        <14>     3/12/97    FA      don't yield too often
        <13>     3/11/97    FA      typo in errors
        <12>     3/10/97    FA      Also implemented using URLOpen with filespec
        <11>      3/9/97    FA      let event routine handle termination
        <10>      3/7/97    FA      Show all events in notification
         <9>      3/5/97    FA      more debug. SaveBufferToFile is now called from the
                                    Notification routine (and yields). The progress dialog is
                                    called regularly from the main routine to update itself.
         <8>      3/4/97    FA      pass filename to progress dialog
         <7>      3/3/97    FA      Adding Post Processing
         <6>      3/3/97    FA      Abort if error
         <5>     2/27/97    FA      release buffers after use
         <4>     2/26/97    FA      problems with CreateUniqueName
         <3>     2/25/97    FA      getting file type and creator. Code review changes.
         <2>     2/20/97    FA      no more downloadFlags, only openFlags
         <1>     2/20/97    FA      first checked in
*/

#include "Subwoofer.h"
#include "SubwooferProgress.h"
#include "StuffItEngineLib.h"
#include "StuffItEngineLib_LL.h"

#include <Files.h>
#include <Finder.h>
#include <Threads.h>
#include <string.h>
#include <Memory.h>
#include <Resources.h>

// Constants and Macros

#define kFirstSupportedStuffItVersion		0x0400
inline void FailError(OSErr err) { if ( err ) throw ( err ); }

enum
{
	kDefaultFileCreator		= '????',
	kDefaultFileType		= '????',
	
	kIntervalTime			= 15
};

// Type definitions

typedef struct URLDataStruct
{
	Boolean				done;				// indicates that the download is done or terminated
	
	FSSpec*				fileSpec;			// the file to dowload to
	SInt16				refNum;				// destination file reference number
	URLOpenFlags		openFlags;			// open flags for post processing

	Size				resourceSize;		// total size of the resource to download
	Size				amountDone;			// the total size downloaded so far
	Str255				mimeType;			// MIME type of the resource to download
	SInt16				dataAvailableCount;	// the number of outstanding data available events
	
	ProgressDataStruct	progressData;		// Info for progress dialog
	Boolean				hasCreatedProgress;	// indicates CreateProgress has been called
} URLDataStruct;

// Prototypes

void PrepareFileForDownload(				const char* url,
											FSSpec *file,
											Boolean createFileDeleteCurFile,
											SInt16* refNum);
static OSStatus NotificationCallback(		UInt32 refCon,
											URLReference urlRef,
											URLEvent event,
											const char* changed,
											UInt32 downloadSize);
Boolean SaveBufferToFile(					SInt16 refNum,
											Size *amountDone,
											URLReference urlRef);
void SetTypeAndCreator(						URLReference urlRef,
											const FSSpec* file,
											Str255 mimeType);
void CreateUniqueName(						FSSpec *file);
void MakeFileName(							char* base,
											UInt32 count,
											char* extension,
											StringPtr name);
Boolean HasNewIntervalTime(					UInt32 *oldTick,
											UInt32 interval);
void GetNameFromURL(						const char* url,
											Str63 name);
// (REMOVE THIS FOR SAMPLE CODE)
void ExpandFile(							FSSpec* file,
											URLOpenFlags openFlags);

// Implementation

extern OSStatus URLDownload(				const char* url,
											FSSpec *destination,
											URLOpenFlags openFlags)
{
	OSErr err = noErr;
	
	URLDataStruct data = { false, destination, 0, openFlags, 0, 0, "\p", 0, false };
	URLReference urlRef = nil;
	URLState theState = 0;
		
	try
	{
		// create new URL reference
		err = URLNewReference(url, &urlRef);
		FailError(err);

		// check the parameters
		if ( data.fileSpec == nil )
		{
			FailError(kURLInvalidParameterError);
		}
		else
		{
			PrepareFileForDownload(	url,
									data.fileSpec,
									kURLReplaceExistingFileFlag & openFlags,
									&data.refNum);
		}

		// now start the dowload
		UInt32 eventRegister =	kURLAllEventsFlag;
		err = URLOpen(	urlRef,
						nil,
						openFlags,
						(URLNotifyProcPtr) &NotificationCallback,
						eventRegister,
						(UInt32) &data);
							
		UInt32 lastUpdateTime = TickCount();
		Boolean wasCancelled = false;

		while ( !data.done )
		{
			YieldToAnyThread();
			
			// save buffers if we need to
			if ( data.dataAvailableCount )
				if ( SaveBufferToFile(data.refNum, &data.amountDone, urlRef) )
					data.dataAvailableCount--;

			// if the file is closed, get rid of progress
			if ( !data.refNum )
				DisposeProgress(&data.progressData);
		}

		// just in case we have a race condition
		DisposeProgress(&data.progressData);

		// Check for errors
		OSStatus error;
		err = URLGetError(urlRef, &error);
		FailError(err);
		FailError(error);
		
		// Expand the file if the client requested it
		if ( kURLExpandFileFlag & openFlags )
			ExpandFile(destination, openFlags);
	}
	catch ( OSErr& _error )
	{
		err = _error;
	}
	
	if ( urlRef )
			(void) URLDisposeReference(urlRef);

	return err;
}

void PrepareFileForDownload(				const char* url,
											FSSpec *file,
											Boolean createFileDeleteCurFile,
											SInt16* refNum)
{
	OSErr err = noErr;
	Boolean fileTouched = false;
	
	*refNum = 0; // so we don't think the file is opened
	
	try
	{
		// if the file name is empty, use the URL
		if ( file->name[0] == 0 )
			GetNameFromURL(url, file->name);
			
		FInfo finderInfo;

		err = FSpGetFInfo(file, &finderInfo);
		if ( err != fnfErr )
			FailError(err); // something serious happened
		
		// if the file exists...
		if ( err == noErr )
		{
			if ( createFileDeleteCurFile ) // we may delete this file...
			{
				err = FSpDelete(file);
				FailError(err);
			}
			else
			{
				CreateUniqueName(file);
			}
		}

		// now we can safely create the file and open it
		fileTouched = true;
		
		err = FSpCreate(file, kDefaultFileCreator, kDefaultFileType, smSystemScript);
		FailError(err);
		err = FSpOpenDF(file, fsWrPerm, refNum);
		FailError(err);
		err = SetFPos(*refNum, fsFromStart, 0);
		FailError(err);
	}
	catch ( OSErr& _error )
	{
		err = _error;
		
		// close file if opened
		if ( *refNum )
			err = FSClose(*refNum);
		*refNum = 0;
		
		// delete the file if it was touched
		if ( fileTouched )
			err = FSpDelete(file);

		throw;
	}
}

OSStatus NotificationCallback(				UInt32 refCon,
											URLReference urlRef,
											URLEvent event,
											const char* changed,
											UInt32 downloadSize)
{
	OSErr error = noErr;
	
	URLDataStruct* data = (URLDataStruct*) refCon;
	Boolean wasCancelled = false;
	
	switch ( event )
	{
	case kURLInitiatedEvent:
		if ( !data->hasCreatedProgress ) // call this only once
		{
			// Initialize the progress bar info
			error = CreateProgress(&data->progressData, data->openFlags & kURLDisplayProgressFlag, data->fileSpec->name);
			data->hasCreatedProgress = true;
			
			// put up first state
			if ( data->openFlags & kURLDisplayProgressFlag )
			{
				Str255 statusString;
				error = URLGetProperty(urlRef, kURLStatusString, &statusString, sizeof ( Str255 ) );
				// call the progress bar dialog with the new status string
				wasCancelled = DoProgress(&data->progressData, 0, 0, statusString);
			}
		}
		break;
		
	case kURLPercentEvent :
	case kURLPeriodicEvent :
		if ( data->openFlags & kURLDisplayProgressFlag )
			wasCancelled = DoProgress(&data->progressData, downloadSize, data->resourceSize, nil);
		break;
		
	case kURLPropertyChangedEvent :
		if ( strcmp(changed, kURLResourceSize) == 0 )
		{
			error = URLGetProperty(urlRef, kURLResourceSize, &data->resourceSize, sizeof ( UInt32 ) );
		}
		else if ( strcmp(changed, kURLMIMEType) == 0 )
		{
			error = URLGetProperty(urlRef, kURLMIMEType, &data->mimeType, sizeof ( Str255 ) );
		}
		else if ( strcmp(changed, kURLStatusString) == 0 )
		{
			Str255 statusString;
			error = URLGetProperty(urlRef, kURLStatusString, &statusString, sizeof ( Str255 ) );
			
			if ( data->openFlags & kURLDisplayProgressFlag )
			{
				// call the progress bar dialog with the new status string
				wasCancelled = DoProgress(&data->progressData, downloadSize, data->resourceSize, statusString);
			}
		}
		break;
	
	case kURLErrorOccurredEvent:
		break;
	
	case kURLResourceFoundEvent:
	case kURLDownloadingEvent:
		if ( data->openFlags & kURLDisplayProgressFlag )
		{
			wasCancelled = DoProgress(&data->progressData, downloadSize, data->resourceSize, nil);
		}
		break;
		
	case kURLAbortInitiatedEvent:
		if ( data->refNum )
		{
			FSClose(data->refNum);
			data->refNum = 0;
			FSpDelete(data->fileSpec);
		}
		break;

	case kURLDataAvailableEvent :
		data->dataAvailableCount++;
		break;
		
	case kURLDownloadCompleteEvent:
		// close the file and set its type and creator
		if ( data->refNum )
			error = FSClose(data->refNum);
		data->refNum = 0;
		SetTypeAndCreator(urlRef, data->fileSpec, data->mimeType);
		break;

	case kURLCompletedEvent:
		data->done = true;
		break;
				
	default :
		break;
	}
		
	if ( wasCancelled && data->refNum )
		(void) URLAbort(urlRef);
	
	return noErr;
}

Boolean SaveBufferToFile(					SInt16 refNum,
											Size *amountDone,
											URLReference urlRef)
{
	Boolean finished = false;
	OSErr err = noErr;
	
	// get the amount available
	SInt32 dataAvailable;
	err = URLGetDataAvailable(urlRef, &dataAvailable);
	FailError(err);
	
	void* buffer = nil;
	Size bufferSize = 0;
	UInt32 tick = TickCount();
	URLState state;
	
	while ( dataAvailable > 0 && !HasNewIntervalTime(&tick, kIntervalTime) )
	{
		err = URLGetBuffer(urlRef, &buffer, &bufferSize);
		if (err) return finished;
		
		*amountDone += bufferSize;
		
		if ( bufferSize )
		{
			dataAvailable -= bufferSize;
			
			err = FSWrite(refNum, &bufferSize, buffer);
			if (err) return finished;

			err = URLReleaseBuffer(urlRef, buffer);
			if (err) return finished;
		}

		err = URLGetDataAvailable(urlRef, &dataAvailable);
		FailError(err);
	}
	
	if ( !dataAvailable )
		finished = true;

	return finished;
}

void SetTypeAndCreator(						URLReference urlRef,
											const FSSpec* file,
											Str255 mimeType)
{
	OSErr err = noErr;
	
	FInfo fileInfo;
	
	// get the current finder info
	err = FSpGetFInfo(file, &fileInfo);
	FailError(err);
	
	// set the type and creator
	err = URLGetProperty(urlRef, kURLFileType, &fileInfo.fdType, sizeof ( OSType ));
	if ( err )
		fileInfo.fdType = '????';
	err = URLGetProperty(urlRef, kURLFileCreator, &fileInfo.fdCreator, sizeof ( OSType ));
	if ( err )
		fileInfo.fdCreator = '????';
	
	// set the finder info
	err = FSpSetFInfo(file, &fileInfo);
	FailError(err);
}

void CreateUniqueName(						FSSpec *file)
{
	OSErr err = noErr;
	FSSpec temp;
	UInt16 attemptCount = 0;
	
	UInt16 extensionLength = 0;
	const kMaxExtensionStringLength = 6;
	char base[32];
	char extension[kMaxExtensionStringLength]; // the file's extension
	
	try
	{
		err = FSMakeFSSpec(file->vRefNum, file->parID, file->name, &temp);
		FailError(err);
	
		extension[0] = 0;
		
		// get the extension
		UInt16 index = file->name[0];
		while ( index && file->name[index] != '.' )
			index--;
		
		if ( index ) // this means we have an extension
			extensionLength = file->name[0] - index + 1;
	
		// set up the extension if any
		if ( extensionLength > 0 && extensionLength < kMaxExtensionStringLength )
		{
			BlockMoveData(file->name + file->name[0] - extensionLength + 1, extension, extensionLength);
			extension[extensionLength] = 0;
		}

		// set up the base
		BlockMoveData(file->name + 1, base, file->name[0] - extensionLength);
		base[file->name[0] - extensionLength] = 0;
		
		UInt32 count = 1;
		while ( count < 1000 )
		{
			// add a number to the base then add the extension
			MakeFileName(base, count, extension, file->name);
			
			// try to make the FSSpec
			err = FSMakeFSSpec(file->vRefNum, file->parID, file->name, &temp);
			FailError(err);

			count++;
		}
		
		FailError(-1); // this should never happen, but send a generic error in case it does
	}
	catch (OSErr& _error)
	{
		err = _error;

		if (err != fnfErr)
			throw;
	}
}

void MakeFileName(							char* base,
											UInt32 count,
											char* extension,
											StringPtr name)
{
	char numString[4];
	numtostring(count, numString);
	
	if ( strlen(base) + strlen(numString) + strlen(extension) > 31 )
		base[31 - strlen(numString) - strlen(extension)] = 0;
	
	strcpy((char*) name + 1, base);
	strcat((char*) name + 1, numString);
	strcat((char*) name + 1, extension);

	name[0] = strlen((char*) name + 1);
}

// (REMOVE THIS FOR SAMPLE CODE)
void ExpandFile(FSSpec* file, URLOpenFlags openFlags)
{
	long 	magicCookie = 0;
	OSErr	error;
	short	version = 0;
	FSSpec	result;
	
	// check to make sure StuffItEngine exists and is the right version
	error = OpenSITEngine(kUseExternalEngine, &magicCookie);
	
	if (error == noErr)
		version = GetSITEngineVersion(magicCookie);
		
	if (error == noErr && version >= kFirstSupportedStuffItVersion)
	{
	// We have the engine, let's expand!
		short	targetType;
		
		// Go throught the engine as many times as it takes to get to a file that
		// is not recognized.  Most common reason this is necessary is .sit.hqx
		do
		{
			// First, figure out what type of file we have
			targetType = DetermineFileType(file);
			
			if (targetType == ftype_StuffIt)
			{
			// The file is a stuffit archive, so unstuff it
				short 		conflictAction;
				
				if (kURLReplaceExistingFileFlag & openFlags)
					conflictAction = USIT_conflict_replaceExisting;
				else
					conflictAction = USIT_conflict_autoRename;
					
				error = UnStuffFSSpec_Enhanced(magicCookie, file, 
											   nil, 								// unstuff into the same folder
											   &result, 							// we want the resulting file spec
											   kDontPromptForDestination, 			// don't prompt the user
											   kDeleteOriginal, 					// delete the archive when done
									  		   createFolder_Smart, 					// create a new folder if it makes sense
									  		   kSaveComments,						// save any archive comments
									  		   kStopOnAbort,						// if the user says abort, do it
									  		   conflictAction,						// conflict action as set above
									  		   nil,									// no password
									  		   openFlags & kURLDisplayProgressFlag,	// display progress only if requested by client
									  		   kNoCallback,							// no alert callback				
									  		   kNoCallback,							// no status callback
									  		   kNoCallback);						// no periodic callback
									  	
									  	
			}
			else if (targetType != ftype_unknown)
			{
			// The file is some other type that stuffit understands, so expand it
				error = ExpandFSSpec(magicCookie, targetType, file,
									 nil,					// expand file into same folder
									 &result,				// we want the resulting file spec
									 createFolder_Smart,	// create a new folder if it makes sense
									 kDeleteOriginal,		// delete the original file
									 textConvert_Never);	// we have already converted the EOLs
									 
			}
			
			// If the engine does not know about this file type, we assume
			// no expansion is necessary
			
			// If there was no error, copy the resultant file name into our file spec
			// to be eventually passed back to the client
			if (error == noErr)
				BlockMove(result.name, file->name, result.name[0]+1);
		} while (targetType != ftype_unknown);
		
	}
	else
	{
	// All we know how to do without the engine is debinhex, so let's try it
	}
	
	if (magicCookie != 0)
		CloseSITEngine(magicCookie);
	
}

Boolean HasNewIntervalTime(					UInt32 *oldTick,
											UInt32 interval)
{
	UInt32 tick = TickCount();
	
	if ( tick > *oldTick + interval )
	{
		*oldTick = tick;
		return true;
	}
	
	return false;
}

void GetNameFromURL(						const char* url,
											Str63 name)
{
	// get the filename
	UInt16 index = strlen(url);
	
	while ( index && url[index] != '/' )
		index--;
	
	if ( index == 0 )
		index = -1;
	
	UInt16 length = strlen(url) - index - 1;
	if ( length > 31 )
	{
		index = strlen(url) - 32;
		length = 31;
	}
	
	name[0] = length;
	BlockMoveData(url + index + 1, name + 1, length);
}
